import unittest

import torch
import mir_eval.separation
import torch_mir_eval
import numpy as np

from utils import *

BACKPROP = True


class TestBatchedBSS(unittest.TestCase):
    def setUp(self) -> None:
        B = 3
        N = 5
        self.src = np.random.rand(B, N, 44000).astype(np.float32)
        self.est = np.random.rand(B, N, 44000).astype(np.float32)

    # CPU TESTS
    def test_bss_eval_sources_permutation_false_cpu(self):
        src = torch.from_numpy(self.src.copy())
        est = torch.from_numpy(self.est.copy())

        bss_eval_sources = torch_mir_eval.batch_separation.bss_eval_sources
        sdr_b, sir_b, sar_b, _ = bss_eval_sources(src, est, compute_permutation=False)
        sdr, sir, sar, _ = torch_mir_eval.separation.bss_eval_sources(src[0], est[0], False)
        self.assertTrue(torch.allclose(sdr, sdr_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sir, sir_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sar, sar_b[0], rtol=1e-4))

    def test_bss_eval_sources_permutation_true_cpu(self):
        src = torch.from_numpy(self.src.copy())
        est = torch.from_numpy(self.est.copy())

        bss_eval_sources = torch_mir_eval.batch_separation.bss_eval_sources
        sdr_b, sir_b, sar_b, perm_b = bss_eval_sources(src, est, compute_permutation=True)
        sdr, sir, sar, perm = torch_mir_eval.separation.bss_eval_sources(src[0], est[0], True)
        self.assertTrue(torch.allclose(sdr, sdr_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sir, sir_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sar, sar_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(perm.long(), perm_b[0]))

    # GPU TESTS
    @unittest.skipIf(not torch.cuda.is_available(), 'Cuda is not available')
    def test_bss_eval_sources_permutation_false_cuda(self):
        src = torch.from_numpy(self.src.copy()).cuda()
        est = torch.from_numpy(self.est.copy()).cuda()

        bss_eval_sources = torch_mir_eval.batch_separation.bss_eval_sources
        sdr_b, sir_b, sar_b, _ = bss_eval_sources(src, est, compute_permutation=False)
        sdr, sir, sar, _ = torch_mir_eval.separation.bss_eval_sources(src[0], est[0], False)
        self.assertTrue(torch.allclose(sdr, sdr_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sir, sir_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sar, sar_b[0], rtol=1e-4))


    @unittest.skipIf(not torch.cuda.is_available(), 'Cuda is not available')
    def test_bss_eval_sources_permutation_true_cuda(self):
        src = torch.from_numpy(self.src.copy()).cuda()
        est = torch.from_numpy(self.est.copy()).cuda()

        bss_eval_sources = torch_mir_eval.batch_separation.bss_eval_sources
        sdr_b, sir_b, sar_b, perm_b = bss_eval_sources(src, est, compute_permutation=True)
        sdr, sir, sar, perm = torch_mir_eval.separation.bss_eval_sources(src[0], est[0], True)
        self.assertTrue(torch.allclose(sdr, sdr_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sir, sir_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(sar, sar_b[0], rtol=1e-4))
        self.assertTrue(torch.allclose(perm.long(), perm_b[0]))

    @unittest.skipIf(BACKPROP is False, 'System non backpropagable')
    def test_bss_eval_gradient_flow(self):
        with torch.autograd.detect_anomaly():
            src = torch.from_numpy(self.src.copy())
            est = torch.from_numpy(self.est.copy()).requires_grad_()

            bss_eval_sources = torch_mir_eval.batch_separation.bss_eval_sources
            sdr_b, sir_b, sar_b, perm_b = bss_eval_sources(src, est, compute_permutation=False)
            scalar = sdr_b.mean()
            scalar.backward()
            self.assertTrue(est.grad is not None)



import torch


def toeplitz(c, r=None):
    """
    Construct a Toeplitz matrix.

    The Toeplitz matrix has constant diagonals, with c as its first column
    and r as its first row. If r is not given, ``r == c`` is
    assumed.
    Inspired from scipy.linalg.toeplitz


    Parameters
    ----------
    c : tensor_like
        First column of the matrix.  Whatever the actual shape of `c`, it
        will be converted to a 1-D array.
    r : tensor_like, optional
        First row of the matrix. If None, ``r = c`` is assumed.
        r[0] is ignored; the first row of the returned matrix is
        ``[c[0], r[1:]]``.  Whatever the actual shape of `r`, it will be
        converted to a 1-D array.

    Returns
    -------
    A : (len(c), len(r)) torch.Tensor
        The Toeplitz matrix. Dtype is the same as ``(c[0] + r[0]).dtype``.

    Examples
    --------
    >>> toeplitz([1,2,3], [1,4,5,6])
    tensor([[1, 4, 5, 6],
           [2, 1, 4, 5],
           [3, 2, 1, 4]])

    """
    c = torch.as_tensor(c).view(-1)
    r = c if r is None else torch.as_tensor(r).view(-1)
    vals = torch.cat([torch.flip(c, dims=(0,)), r[1:]])
    toep = torch.stack([
        torch.roll(vals, shifts=i, dims=0)[-len(r):] for i in range(len(c))
    ])
    return toep


def batch_toeplitz(c, r=None):
    """
    Construct a batch Toeplitz matrix.

    The Toeplitz matrix has constant diagonals, with c as its first column
    and r as its first row. If r is not given, ``r == c`` is
    assumed.
    Inspired from scipy.linalg.toeplitz

    Parameters
    ----------
    c : tensor_like
        First column of the matrix.  Whatever the actual shape of `c`, it
        will be converted to a batched 1-D array.
    r : tensor_like, optional
        First row of the matrix. If None, ``r = c`` is assumed.
        r[0] is ignored; the first row of the returned matrix is
        ``[c[:, 0], r[:, 1:]]``.  Whatever the actual shape of `r`, it will be
        converted to a batched 1-D array.

    Returns
    -------
    A : (batch, len(c), len(r)) torch.Tensor
        The Toeplitz matrix. Dtype is the same as ``(c[0] + r[0]).dtype``.

    Examples
    --------
    >>> import torch
    >>> batch_toeplitz(torch.randn(2, 3), torch.randn(2, 4))
    """
    c = torch.as_tensor(c).view(c.shape[0], -1)
    r = c if r is None else torch.as_tensor(r).view(r.shape[0], -1)
    assert c.shape[0] == r.shape[0], "Batch dimension shoudl agree between r and c."

    vals = torch.cat([torch.flip(c, dims=(1,)), r[:, 1:]], dim=1)
    toep = torch.stack([
        torch.roll(vals, shifts=i, dims=1)[:, -r.shape[1]:] for i in range(c.shape[1])
    ], dim=1)
    return toep



import unittest

import torch
from torch.testing import assert_allclose
from scipy.linalg import toeplitz as np_toeplitz
from torch_mir_eval.toeplitz import toeplitz, batch_toeplitz


class TestToeplitz(unittest.TestCase):
    def test_toeplitz_nor(self):
        c = [1, 2, 3, 5, 4]
        np_toep = np_toeplitz(c)
        torch_toep = toeplitz(c)
        assert_allclose(torch.from_numpy(np_toep), torch_toep)

    def test_toeplitz(self):
        c, r = [1, 2, 3], [4, 3, 1, 2]
        np_toep = np_toeplitz(c, r=r)
        torch_toep = toeplitz(c, r=r)
        assert_allclose(torch.from_numpy(np_toep), torch_toep)

    def test_batch_toeplitz_nor(self):
        bc = torch.randn(3, 8)
        indirect_toep = torch.stack(
            [toeplitz(c) for c in bc],
            dim=0
        )
        direct_toep = batch_toeplitz(bc)
        assert_allclose(indirect_toep, direct_toep)

    def test_batch_toeplitz(self):
        bc = torch.randn(3, 8)
        br = torch.randn(3, 4)
        indirect_toep = torch.stack(
            [toeplitz(c, r) for c, r in zip(bc, br)],
            dim=0
        )
        direct_toep = batch_toeplitz(bc, br)
        assert_allclose(indirect_toep, direct_toep)

